#pragma once

#include "../utils/utils.h"
#include "../debug/evemon.h"
#include "../debug/exception.h"
#include "../utils/singleton.h"

#include "exception.h"

#include <stdio.h>
#include <string.h>

#include <string>
#include <unordered_map>

/* types manager singleton class */
class types_manager : public singleton<types_manager>
{
	friend class singleton<types_manager>;

public:

	// register C to translated type conversion
	void register_type(const std::string& ctype, const std::string& ttype)
	{
		_debug("registering type \"%s\" as \"%s\"", ctype.c_str(), ttype.c_str());

		this->m_forward.emplace(std::make_pair(ctype, ttype));
		this->m_inverse.emplace(std::make_pair(ttype, ctype));
	}

	// translate ctype
	std::string get_name_from_type(const std::string& ctype) const
	{
		auto it = this->m_forward.find(ctype);

		if (it == this->m_forward.end())
			throw_exception(cannot_resolve_typename_exception, ctype);

		return it->second;
	}

	// inverse translate
	std::string get_type_from_name(const std::string& ttype) const
	{
		auto it = this->m_inverse.find(ttype);

		if (it == this->m_inverse.end())
			throw_exception(cannot_resolve_typename_exception, ttype);

		return it->second;
	}

private:

	// ctor
	types_manager(void) {}

	// lists for conversion
	std::unordered_map<std::string, std::string> m_forward, m_inverse;
};

/* type_auto_register class */
class type_auto_register
{
public:
	type_auto_register(const std::string& ctype, const std::string& ttype)
	{
		get_instance<types_manager>()->register_type(ctype, ttype);
	}
};

// translate type to name
template<typename type> std::string get_name_from_type(void)
{
	return get_instance<types_manager>()->get_name_from_type(typeid(type).name());
}

template<typename type> std::string get_name_from_type(const type* p)
{
	return get_name_from_type<type>();
}

// return true if types matches
template<typename type> bool is_type_name(const std::string& name)
{
	return _strcmpi(get_name_from_type<type>().c_str(), name.c_str()) == 0;
}

// translate name to type
static std::string get_type_from_name(const std::string& name)
{
	return get_instance<types_manager>()->get_type_from_name(name);
}

// register typename
#define REGISTER_TYPENAME_EX(class, string, var)	type_auto_register register_##var(typeid(class).name(), string);

#define REGISTER_TYPENAME(type)						REGISTER_TYPENAME_EX(type, #type, type);