#include <stdio.h>

#include "core/utils/utils.h"
#include "core/instancers/instancer.h"
#include "core/debug/evemon.h"

#include "msg.h"
#include "example1.h"

// do something with pointer
void process_global(std::shared_ptr<msg_base> p)
{
	if (__ISNULLPTR(p))
		return;

	p->print();
}

// main program
int main(int argc, char** argv)
{
	try
	{
		// local example
		printf("local instancer example:\r\n");

		{
			example1 obj;

			{
				auto p = obj.create_instance(get_name_from_type<msg_type_a>());

				obj.process_local(p);
			}

			{
				auto p = obj.create_instance(get_name_from_type<msg_type_b>());

				obj.process_local(p);
			}
		}

		printf("\r\n");

		// global example #1
		printf("global instancer example:\r\n");

		{
			auto p = create_instance<msg_base>(get_name_from_type<msg_type_a>());

			process_global(p);
		}

		{
			auto p = create_instance<msg_base>(get_name_from_type<msg_type_b>());

			process_global(p);
		}

		// return successfuly
		return 0;
	}
	catch (exception_base& e)
	{
		printf("error: %s\r\n", e.to_string().c_str());

		_critical("%s", e.to_string().c_str());
	}
	catch (std::exception& e)
	{
		printf("error: %s\r\n", e.what());

		_critical("%s", e.what());
	}
	catch (...)
	{
		printf("unknown error!\r\n");

		_critical("unknown error!");
	}

	// signal error
	return 1;
}